#ifndef MAPCLASS_H
#define MAPCLASS_H

#include <vector>

#define HEX_Z_OFFSET 1.495f
#define HEX_X_OFFSET 1.55f

#define TYPE_NORMAL  0
#define TYPE_BLOCKED 1
#define TYPE_START   2
#define TYPE_FINISH  3

#define LIST_NONE   0
#define LIST_OPEN   1
#define LIST_CLOSED 2
#define LIST_PATH   3

#define PATH_OK            0
#define PATH_ERR_START_END 1
#define PATH_ERR_NO_PATH   2

#define STEP_START 0
#define STEP_SEEK  1
#define STEP_PATH  2

#define STEP_CONTINUE 1001


typedef struct HexStruct
	{
  int Type;
  int Previous;
		int List;
		float F,G,H;
	}HexStruct;

typedef struct HexPosition
	{
	 int Row,Col;
	}HexPosition;
	
typedef struct HexCoordinate
 {
  float x,z;
 }HexCoordinate;

class MapObj
	{
	 public:
   MapObj();
			void Init(int Rows, int Cols); // Populate map vector
			void Reset(); // Clear map attributes
			void Render(); // Draw the map
   void Render(bool Select); // Draw the map with cheap selection trick
			void SetHexType(int Num, int Type); // Manage hex attributes
			HexPosition GetPos(int HexNum);  // Get the row and column position of a hex
			HexCoordinate GetCoord(int HexNum); // Get the coordinates of a hex
			int GetNeigbour(int HexNum, int Which); // Find adjacent hexes
			int FindPath(); // Do the path finding
			int FindPathStep(bool Init); // Do the path finding in single steps
			
			std::vector<int> Path; // Contains the path hex indices

	 private:
   std::vector<HexStruct> Hex; // Map array
   std::vector<HexStruct>::iterator HexItr; // Map Array iterator
   
   float CalcH(int SrcHex, int TgtHex); // Calculates the distance from the specified hex to the end point
   int Start,End; // Indices of the start and end hexes
			int HexCount,RowPitch; // Map size parameters
	};

#endif
