// This code will be executed by the OpenGL base app's render thread
#include "Framework.h" 
#include "tga.h"
#include "MapClass.h"
#include "BitmapFontClass.h"
#include <iostream>
#include <string>
#include <ctime>

#define MODE_NORMAL      0
#define MODE_BLOCKING    1
#define MODE_START_POINT 2
#define MODE_END_POINT   3

#define STEP_DELAY 0.2

GLuint LoadTexture(char *TexName);

using namespace std;

void Render(void)
{
  MapObj Map;
  CBitmapFont Fnt;
  GLuint TexID;
  int Mode=MODE_START_POINT; // Mode for setting hex types
  int Result;
  double CurTime,TgtTime,LastFrame,TBMVal;
  char MRed,MGreen;
  int Selected;
  string PickMsg,StepMsg,Message; // Message strings
  float ScreenRatio;
  float X,Y,Z;                    // Observer coodinates
  // Flags for options
  bool Clicked=false,ShowSelect=false,ShowSteps=false,Stepping=false;

  // Create the map
  Map.Init(15,21);

  // Set the background to black
  glClearColor(0,0,0,0);
  glClear(GL_COLOR_BUFFER_BIT);

  // Set up view
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glViewport(0,0,800,600);
  ScreenRatio=600.0f/800.0f;
  glFrustum(-1.0f,1.0f,-1.0f*ScreenRatio,1.0f*ScreenRatio,1.0f,100.0f);
  glMatrixMode(GL_MODELVIEW);
  
  X=-15.5f;
  Y=-14.0f;
  Z=-15.0f;

  // Load button image
  TexID=LoadTexture("buttons.tga");
  if(TexID==-1)
  {
    std::cout<<"Error - Unable to load texture buttons.tga\n";
    RunLevel=0;
  }

  // Load font
  if(!Fnt.Load("Arial.bff"))
  {
    std::cout<<"Error - Unable to load font Arial.bff\n";
    RunLevel=0;
  }

  // Set up the message strings
  Message="Pathfinding Demo - A* on a hex based map. Use the buttons to set up the map and find a route.";
  PickMsg="Show Picking OFF";
  StepMsg="Slow Step Pathfinding OFF";

  // This loop will run until Esc is pressed
  while(RunLevel)
  {
    if(Keys[VK_ESCAPE]) // Esc Key
      RunLevel=0;

    // Get time values for this frame
    LastFrame=CurTime;
    CurTime=glfwGetTime();
    TBMVal=CurTime-LastFrame;

    // Clear last frame
    glClear(GL_COLOR_BUFFER_BIT);

    // Set up view
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glFrustum(-1.0f,1.0f,-1.0f*ScreenRatio,1.0f*ScreenRatio,1.0f,100.0f);
    glMatrixMode(GL_MODELVIEW);

    glRotatef(75.0f,1.0f,0.0f,0.0f);
    glTranslatef(X,Y,Z);

    glDisable(GL_TEXTURE_2D);
    glDisable(GL_BLEND);

    // Handle kestrokes
    if(Keys[GLFW_KEY_F1]) // Toggle visible picking
    {
      if(ShowSelect)
      {
        ShowSelect=false;
        PickMsg="Show Picking OFF";
      }
      else
      {
        ShowSelect=true;
        PickMsg="Show Picking ON";
      }
      Keys[GLFW_KEY_F1]=false;
    }

    if(Keys[GLFW_KEY_F2]) // Toggle slow pathfinding
    {
      if(ShowSteps)
      {
        ShowSteps=false;
        StepMsg="Slow Step Pathfinding OFF";
      }
      else
      {
        ShowSteps=true;
        StepMsg="Slow Step Pathfinding ON";
      }
      Keys[GLFW_KEY_F2]=false;
    }
    
    // Handle observer movement
    if(Keys[GLFW_KEY_UP])
     Z+=(TBMVal*5.0f);
       
    if(Keys[GLFW_KEY_DOWN])
     Z-=(TBMVal*5.0f);
     
    if(Keys[GLFW_KEY_LEFT])
     X+=(TBMVal*5.0f);
     
    if(Keys[GLFW_KEY_RIGHT])
     X-=(TBMVal*5.0f);
     
    if(Keys[GLFW_KEY_PAGEUP])
     {
      Y+=(TBMVal*5.0f);
      Z+=(TBMVal*0.5f);
     }
     
     if(Keys[GLFW_KEY_PAGEDOWN])
     {
      Y-=(TBMVal*5.0f);
      Z-=(TBMVal*0.5f);
     }    
     
     if(Keys[GLFW_KEY_F12])
     {
      X=-15.5f;
      Y=-14.0f;
      Z=-15.0f;     
     }

    // Clear clicked flag on mouse release
    if(!Mouse.Mleft && !Mouse.Mright)
      Clicked=false;

    // Handle mouse input
    if(Mouse.Mleft)
    {
      if(Mouse.My<25)
      {
        if(Mouse.Mx<25)
        {
          //Nothing to click
        }
        else if(Mouse.Mx<175) // Start point button
        {
          Mode=MODE_START_POINT;
          Message="Select Start Point";
        }
        else if(Mouse.Mx<325) // End point button
        {
          Mode=MODE_END_POINT;
          Message="Select End Point";
        }
        else if(Mouse.Mx<475) // Non walkable button
        {
          Mode=MODE_BLOCKING;
          Message="Set non-walkable hexes, right click to erase.";
        }
        else if(Mouse.Mx<625) // Find route button
        {
          if(ShowSteps) // Step-by-step pathfinding
          {
            if(Stepping==false)
            {
              Stepping=true;
              TgtTime=CurTime+STEP_DELAY;
              Map.FindPathStep(true);
            }
          }
          else // Normal pathfinding
          {
            Result=Map.FindPath();
          }

          // Set messages for return values
          if(Result!=PATH_OK || Result!=STEP_CONTINUE)
          {
            switch(Result)
            {
            case PATH_ERR_NO_PATH:
              Message="No path could be found!";
              break;

            case PATH_ERR_START_END:
              Message="You need a start and end point to create a path!";
              break;
            }
          }
          else
          {
            Message="Path found.";
          }

        }
        else if(Mouse.Mx<775) // Reset map button
        {
          Map.Reset();
          Message="Map reset.";
        }
      }
      else // Not a button click, check for hex clicks
      {
        Clicked=true;
        Map.Render(true); // Render map with select enabled

        // Get the red value of the pixel under the mouse pointer
        glReadPixels(Mouse.Mx,600-Mouse.My,1,1,GL_RED,GL_BYTE,&MRed);

        if(MRed) // We got something, check for a green value
        {
         glReadPixels(Mouse.Mx,600-Mouse.My,1,1,GL_GREEN,GL_BYTE,&MGreen);
       
       // Calculate the selected hex's index
       Selected=MGreen*126;
       Selected+=(MRed-1);
       
       // Set the selected hex's type
         Map.SetHexType(Selected,Mode);
        }
      }
    }

    // Erase hex properties
    if(Mouse.Mright)
    {
      Clicked=true;
      Map.Render(true); // Render map with selection enabled

      // Get the red value of the pixel under the mouse
      glReadPixels(Mouse.Mx,600-Mouse.My,1,1,GL_RED,GL_BYTE,&MRed);
      
      if(MRed) // We got something, check for a green value
      {
       glReadPixels(Mouse.Mx,600-Mouse.My,1,1,GL_GREEN,GL_BYTE,&MGreen);
       
       // Calculate the selected hex's index
       Selected=MGreen*127;
       Selected+=(MRed-1);
       
       // Reset the selected hex's type
       Map.SetHexType(Selected,TYPE_NORMAL);
      }

    }

    // Do slow stepping if required
    if(Stepping)
    {
      if(CurTime>TgtTime) // Check to see if we need to do the next step
      {
        TgtTime=CurTime+STEP_DELAY; // Set the next step's time value

        Result=Map.FindPathStep(false); // Call the pathfinder

        // Check return, anything other than STEP_CONTINUE means it's finished
        if(Result!=STEP_CONTINUE)
        {
          Stepping=false;

          switch(Result)
          {
          case PATH_OK:
            Message="Path Found";
            break;

          case PATH_ERR_NO_PATH:
            Message="No path could be found!";
            break;
          }
        }
      }
    }

    // Render map unless there has been a click and show picking is ON
    if(Clicked && ShowSelect)
    {
      // Leave selection render visible
    }
    else
    {
      Map.Render();
    }


    // Draw controls
    Fnt.SetScreen(800,600);
    glColor3f(1.0f,1.0f,1.0f);

    glEnable(GL_TEXTURE_2D);
    glBindTexture(GL_TEXTURE_2D,TexID);

    // Draw buttons
    glBegin(GL_QUADS);
    glTexCoord2f(0.0f,0.5f); glVertex2i(25,600);
    glTexCoord2f(1.0f,0.5f); glVertex2i(175,600);
    glTexCoord2f(1.0f,0.7f); glVertex2f(175,575);
    glTexCoord2f(0.0f,0.7f); glVertex2f(25,575);

    glTexCoord2f(0.0f,0.75f); glVertex2i(175,600);
    glTexCoord2f(1.0f,0.75f); glVertex2i(325,600);
    glTexCoord2f(1.0f,0.95f);  glVertex2f(325,575);
    glTexCoord2f(0.0f,0.95f);  glVertex2f(175,575);

    glTexCoord2f(0.0f,0.25f); glVertex2i(325,600);
    glTexCoord2f(1.0f,0.25f); glVertex2i(475,600);
    glTexCoord2f(1.0f,0.45f);  glVertex2f(475,575);
    glTexCoord2f(0.0f,0.45f);  glVertex2f(325,575);

    glTexCoord2f(0.0f,0.0f);  glVertex2i(475,600);
    glTexCoord2f(1.0f,0.0f);  glVertex2i(625,600);
    glTexCoord2f(1.0f,0.2f); glVertex2f(625,575);
    glTexCoord2f(0.0f,0.2f); glVertex2f(475,575);

    glTexCoord2f(0.0f,0.0f);  glVertex2i(625,600);
    glTexCoord2f(1.0f,0.0f);  glVertex2i(775,600);
    glTexCoord2f(1.0f,0.2f); glVertex2f(775,575);
    glTexCoord2f(0.0f,0.2f); glVertex2f(625,575);
    glEnd();

    // Draw button text
    Fnt.SetColor(0.0f,0.0f,0.0f);
    Fnt.Select();

    Fnt.Print("Set Start Point",55,578);
    Fnt.Print("Set End Point",205,578);
    Fnt.Print("Set Non-Walkable",345,578);
    Fnt.Print("Find Route",515,578);
    Fnt.Print("Reset Map",665,578);

    Fnt.SetColor(1.0f,1.0f,1.0f);
    Fnt.SetBlend();

    // Highlight current button
    switch(Mode)
    {
    case MODE_START_POINT:
      Fnt.Print("Set Start Point",55,578);
      break;

    case MODE_END_POINT:
      Fnt.Print("Set End Point",205,578);
      break;

    case MODE_BLOCKING:
      Fnt.Print("Set Non-Walkable",345,578);
      break;
    }
    
    // Fade bottom of screen
    glDisable(GL_TEXTURE_2D);
    glBegin(GL_QUADS);
     glColor4f(0.0f,0.0f,0.0f,0.9f); 
     glVertex2i(  0,0); 
     glVertex2i(800,0);
     glVertex2i(800,60);
     glVertex2i(  0,60);
     glColor4f(0.0f,0.0f,0.0f,0.9f); glVertex2i(  0,60); 
     glColor4f(0.0f,0.0f,0.0f,0.9f); glVertex2i(800,60);
     glColor4f(0.0f,0.0f,0.0f,0.0f); glVertex2i(800,80);
     glColor4f(0.0f,0.0f,0.0f,0.0f); glVertex2i(  0,80);
    glEnd();    
    glEnable(GL_TEXTURE_2D);

    // Draw status messages
    glColor3f(1.0f,1.0f,1.0f);
    Fnt.Print((char*)PickMsg.c_str(),15,40);
    Fnt.Print((char*)StepMsg.c_str(),15,23);
    Fnt.Print((char*)Message.c_str(),15,5);

    // We're using double buffers, so we need to swap to see our stuff
    FlipBuffers();
  }

}


// Texture loader
GLuint LoadTexture(char *TexName)
{
  TGAImg Img;        // Image loader
  GLuint Texture;

  // Load our Texture
  if(Img.Load(TexName)!=IMG_OK)
    return -1;

  glGenTextures(1,&Texture);            // Allocate space for texture
  glBindTexture(GL_TEXTURE_2D,Texture); // Set our Tex handle as current

  // Create the texture
  if(Img.GetBPP()==24)
    glTexImage2D(GL_TEXTURE_2D,0,3,Img.GetWidth(),Img.GetHeight(),0,GL_RGB,GL_UNSIGNED_BYTE,Img.GetImg());
  else if(Img.GetBPP()==32)
    glTexImage2D(GL_TEXTURE_2D,0,4,Img.GetWidth(),Img.GetHeight(),0,GL_RGBA,GL_UNSIGNED_BYTE,Img.GetImg());
  else
    return -1;

  // Specify filtering and edge actions
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_S,GL_CLAMP);
  glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_WRAP_T,GL_CLAMP);

  return Texture;
}

